package org.deephacks.tools4j.config.internal.core.runtime;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.deephacks.tools4j.config.Config;
import org.deephacks.tools4j.config.Id;
import org.deephacks.tools4j.config.internal.core.runtime.ClassIntrospector.FieldWrap;
import org.deephacks.tools4j.config.model.Bean;
import org.deephacks.tools4j.config.model.Bean.BeanId;
import org.deephacks.tools4j.config.model.Schema;
import org.deephacks.tools4j.config.spi.Conversion;
import org.deephacks.tools4j.config.spi.Conversion.Converter;

public class ObjectToBeanConverter implements Converter<Object, Bean> {
    Conversion conversion = Conversion.get();

    @Override
    public Bean convert(Object source, Class<? extends Bean> specificType) {
        ClassIntrospector i = new ClassIntrospector(source.getClass());
        Bean bean = Bean.create(getBeanId(source));

        for (FieldWrap<Config> field : i.getFieldList(Config.class)) {
            Object value = field.getValue(source);
            if (value == null) {
                continue;
            }
            addProperty(value, bean, field);
        }

        Schema schema = conversion.convert(source.getClass(), Schema.class);
        bean.set(schema);
        return bean;
    }

    @SuppressWarnings("unchecked")
    private void addProperty(Object value, Bean bean, FieldWrap<Config> fieldwrap) {
        Config field = fieldwrap.getAnnotation();
        String name = field.name();
        if (name == null || "".equals(name)) {
            name = fieldwrap.getFieldName();
        }
        if (fieldwrap.isCollection()) {
            Class<?> paramClass = fieldwrap.getType();
            Collection<Object> values = (Collection<Object>) value;
            if (paramClass.isAnnotationPresent(Config.class)) {
                for (Object object : values) {
                    bean.addReference(name, getRecursiveBeanId(object));
                }
            } else {
                bean.addProperty(name, conversion.convert(values, String.class));
            }
        } else if (fieldwrap.isMap()) {
            Class<?> paramClass = fieldwrap.getMapParamTypes().get(1);
            Map<String, Object> values = (Map<String, Object>) value;
            if (paramClass.isAnnotationPresent(Config.class)) {
                for (Object object : values.values()) {
                    bean.addReference(name, getRecursiveBeanId(object));
                }
            }
        } else {
            if (value.getClass().isAnnotationPresent(Config.class)) {
                bean.addReference(name, getRecursiveBeanId(value));
            } else {
                String converted = conversion.convert(value, String.class);
                bean.setProperty(name, converted);
            }
        }
    }

    private BeanId getBeanId(Object bean) {
        Config config = bean.getClass().getAnnotation(Config.class);
        if (config == null) {
            throw new IllegalArgumentException("Bean [" + bean
                    + "] does not have a @Config annotation.");
        }
        String schemaName = config.name();
        if (schemaName == null || "".equals(schemaName.trim())) {
            schemaName = bean.getClass().getName();
        }
        ClassIntrospector i = new ClassIntrospector(bean.getClass());
        List<FieldWrap<Id>> ids = i.getFieldList(Id.class);
        if (ids == null || ids.size() != 1) {
            // a singleton
            return BeanId.createSingleton(schemaName);
        }
        FieldWrap<Id> id = ids.get(0);
        BeanId targetId = null;
        Object idValue = id.getValue(bean);
        if (idValue == null) {
            throw new IllegalArgumentException("Cannot convert an objects that have emtpty id ["
                    + bean + "]");
        }
        targetId = BeanId.create(id.getValue(bean).toString(), schemaName);
        return targetId;
    }

    private BeanId getRecursiveBeanId(Object bean) {
        BeanId targetId = getBeanId(bean);
        Bean targetBean = conversion.convert(bean, Bean.class);
        targetBean.set(conversion.convert(bean.getClass(), Schema.class));
        targetId.setBean(targetBean);
        return targetId;
    }
}